from fastapi import FastAPI, status

from .forwarding import get_fwd_anything_objs, get_fwd_openai_style_objs

app = FastAPI(title="openai_forward", version="0.4")


@app.get(
    "/healthz",
    summary="Perform a Health Check",
    response_description="Return HTTP Status Code 200 (OK)",
    status_code=status.HTTP_200_OK,
)
def healthz():
    return "OK"


add_route = lambda obj: app.add_route(
    obj.ROUTE_PREFIX + "{api_path:path}",
    obj.reverse_proxy,
    methods=["GET", "POST", "PUT", "DELETE", "OPTIONS", "HEAD", "PATCH", "TRACE"],
)

[add_route(obj) for obj in get_fwd_openai_style_objs()]
[add_route(obj) for obj in get_fwd_anything_objs()]



from httpx._decoders import LineDecoder, TextChunker, TextDecoder


def iter_text(iter_tytes: list):
    decoder = TextDecoder("utf-8")
    chunker = TextChunker()
    for byte_content in iter_tytes:
        text_content = decoder.decode(byte_content)
        for chunk in chunker.decode(text_content):
            yield chunk
    text_content = decoder.flush()
    for chunk in chunker.decode(text_content):
        yield chunk
    for chunk in chunker.flush():
        yield chunk


def parse_to_lines(iter_bytes: list) -> list:
    decoder = LineDecoder()
    lines = []
    for text in iter_text(iter_bytes):
        lines.extend(decoder.decode(text))
    lines.extend(decoder.flush())
    return lines



import os

from ..config import setting_log
from ..helper import env2list, format_route_prefix

ENV_VAR_SEP = ","
OPENAI_BASE_URL = env2list("OPENAI_BASE_URL", sep=ENV_VAR_SEP) or [
    "https://api.openai.com"
]

OPENAI_ROUTE_PREFIX = [
    format_route_prefix(i) for i in env2list("OPENAI_ROUTE_PREFIX", sep=ENV_VAR_SEP)
] or ['/']

EXTRA_BASE_URL = env2list("EXTRA_BASE_URL", sep=ENV_VAR_SEP)
EXTRA_ROUTE_PREFIX = [
    format_route_prefix(i) for i in env2list("EXTRA_ROUTE_PREFIX", sep=ENV_VAR_SEP)
]

LOG_CHAT = os.environ.get("LOG_CHAT", "False").strip().lower() == "true"
if LOG_CHAT:
    setting_log(save_file=False)

IP_WHITELIST = env2list("IP_WHITELIST", sep=ENV_VAR_SEP)
IP_BLACKLIST = env2list("IP_BLACKLIST", sep=ENV_VAR_SEP)

OPENAI_API_KEY = env2list("OPENAI_API_KEY", sep=ENV_VAR_SEP)
FWD_KEY = env2list("FORWARD_KEY", sep=ENV_VAR_SEP)

PROXY = os.environ.get("PROXY", "").strip()
PROXY = PROXY if PROXY else None



import time
import uuid
from typing import List

import orjson
from fastapi import Request
from loguru import logger
from orjson import JSONDecodeError

from .decode import parse_to_lines


class ChatSaver:
    def __init__(self):
        self.logger = logger.bind(chat=True)

    @staticmethod
    async def parse_payload(request: Request):
        uid = uuid.uuid4().__str__()
        payload = await request.json()
        msgs = payload["messages"]
        model = payload["model"]
        content = {
            "messages": [{msg["role"]: msg["content"]} for msg in msgs],
            "model": model,
            "forwarded-for": request.headers.get("x-forwarded-for") or "",
            "uid": uid,
            "datetime": time.strftime("%Y-%m-%d %H:%M:%S", time.localtime()),
        }
        return content

    def parse_iter_bytes(self, byte_list: List[bytes]):
        txt_lines = parse_to_lines(byte_list)

        start_line = txt_lines[0]
        target_info = dict()
        start_token = "data: "
        start_token_len = len(start_token)
        if start_line.startswith(start_token):
            stream = True
            start_line = orjson.loads(start_line[start_token_len:])
            msg = start_line["choices"][0]["delta"]
        else:
            stream = False
            start_line = orjson.loads("".join(txt_lines))
            msg = start_line["choices"][0]["message"]

        target_info["created"] = start_line["created"]
        target_info["id"] = start_line["id"]
        target_info["model"] = start_line["model"]
        target_info["role"] = msg["role"]
        target_info["content"] = msg.get("content", "")

        if not stream:
            return target_info

        # loop for stream
        for line in txt_lines[1:]:
            if line.startswith(start_token):
                target_info["content"] += self._parse_one_line(line[start_token_len:])
        return target_info

    @staticmethod
    def _parse_one_line(line: str):
        try:
            line_dict = orjson.loads(line)
            return line_dict["choices"][0]["delta"]["content"]
        except JSONDecodeError:
            return ""
        except KeyError:
            return ""

    def add_chat(self, chat_info: dict):
        self.logger.debug(f"{chat_info}")



import traceback
from itertools import cycle
from typing import Any, AsyncGenerator

import httpx
from fastapi import HTTPException, Request, status
from fastapi.responses import StreamingResponse
from loguru import logger

from ..content import ChatSaver, ExtraForwardingSaver, WhisperSaver
from .settings import *


class ForwardingBase:
    BASE_URL = None
    ROUTE_PREFIX = None
    client: httpx.AsyncClient = None
    if IP_BLACKLIST or IP_WHITELIST:
        validate_host = True
    else:
        validate_host = False

    timeout = 600

    if LOG_CHAT:
        extrasaver = ExtraForwardingSaver()

    @staticmethod
    def validate_request_host(ip):
        if IP_WHITELIST and ip not in IP_WHITELIST:
            raise HTTPException(
                status_code=status.HTTP_403_FORBIDDEN,
                detail=f"Forbidden, ip={ip} not in whitelist!",
            )
        if IP_BLACKLIST and ip in IP_BLACKLIST:
            raise HTTPException(
                status_code=status.HTTP_403_FORBIDDEN,
                detail=f"Forbidden, ip={ip} in blacklist!",
            )

    @classmethod
    async def aiter_bytes(
        cls, r: httpx.Response, **kwargs
    ) -> AsyncGenerator[bytes, Any]:
        bytes_ = b""
        async for chunk in r.aiter_bytes():
            bytes_ += chunk
            yield chunk
        cls.extrasaver.add_log(bytes_)

    async def try_send(self, req: httpx.Request, request: Request):
        try:
            r = await self.client.send(req, stream=True)
            return r
        except (httpx.ConnectError, httpx.ConnectTimeout) as e:
            error_info = (
                f"{type(e)}: {e} | "
                f"Please check if host={request.client.host} can access [{self.BASE_URL}] successfully?"
            )
            logger.error(error_info)
            raise HTTPException(
                status_code=status.HTTP_504_GATEWAY_TIMEOUT, detail=error_info
            )
        except Exception as e:
            logger.exception(f"{type(e)}:")
            raise HTTPException(
                status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, detail=e
            )

    def prepare_client(self, request: Request):
        assert self.BASE_URL is not None
        assert self.ROUTE_PREFIX is not None
        if self.validate_host:
            ip = request.headers.get("x-forwarded-for") or ""
            self.validate_request_host(ip)

        url_path = request.url.path
        prefix_index = 0 if self.ROUTE_PREFIX == '/' else len(self.ROUTE_PREFIX)

        route_path = url_path[prefix_index:]
        url = httpx.URL(path=route_path, query=request.url.query.encode("utf-8"))
        headers = dict(request.headers)
        auth = headers.pop("authorization", "")
        content_type = headers.pop("content-type", "application/json")
        auth_headers_dict = {"Content-Type": content_type, "Authorization": auth}
        client_config = {
            'auth': auth,
            'headers': auth_headers_dict,
            'url': url,
            'url_path': route_path,
        }

        return client_config

    async def reverse_proxy(self, request: Request):
        assert self.client is not None
        client_config = self.prepare_client(request)

        req = self.client.build_request(
            request.method,
            client_config['url'],
            headers=client_config["headers"],
            content=request.stream(),
            timeout=self.timeout,
        )
        r = await self.try_send(req, request)

        return StreamingResponse(
            self.aiter_bytes(r),
            status_code=r.status_code,
            media_type=r.headers.get("content-type"),
        )


class OpenaiBase(ForwardingBase):
    _cycle_api_key = cycle(OPENAI_API_KEY)
    _no_auth_mode = OPENAI_API_KEY != [] and FWD_KEY == set()

    if LOG_CHAT:
        chatsaver = ChatSaver()
        whispersaver = WhisperSaver()

    async def aiter_bytes(
        self, r: httpx.Response, request: Request, route_path: str, uid: str
    ):
        byte_list = []
        async for chunk in r.aiter_bytes():
            byte_list.append(chunk)
            yield chunk
        await r.aclose()
        try:
            if LOG_CHAT and request.method == "POST":
                if route_path == "/v1/chat/completions":
                    target_info = self.chatsaver.parse_iter_bytes(byte_list)
                    self.chatsaver.add_chat(
                        {target_info["role"]: target_info["content"], "uid": uid}
                    )
                elif route_path.startswith("/v1/audio/"):
                    self.whispersaver.add_log(b"".join([_ for _ in byte_list]))
        except Exception as e:
            logger.debug(f"log chat (not) error:\n{traceback.format_exc()}")

    async def reverse_proxy(self, request: Request):
        client_config = self.prepare_client(request)
        auth_prefix = "Bearer "
        auth = client_config["auth"]
        auth_headers_dict = client_config["headers"]
        url_path = client_config["url_path"]
        if self._no_auth_mode or auth and auth[len(auth_prefix) :] in FWD_KEY:
            auth = auth_prefix + next(self._cycle_api_key)
            auth_headers_dict["Authorization"] = auth

        uid = None
        if LOG_CHAT and request.method == "POST":
            try:
                if url_path == "/v1/chat/completions":
                    chat_info = await self.chatsaver.parse_payload(request)
                    if chat_info:
                        self.chatsaver.add_chat(chat_info)
                        uid = chat_info.get("uid")
            except Exception as e:
                logger.debug(
                    f"log chat error:\nhost:{request.client.host} method:{request.method}: {traceback.format_exc()}"
                )

        req = self.client.build_request(
            request.method,
            client_config['url'],
            headers=auth_headers_dict,
            content=request.stream(),
            timeout=self.timeout,
        )
        r = await self.try_send(req, request)
        aiter_bytes = self.aiter_bytes(r, request, url_path, uid)
        return StreamingResponse(
            aiter_bytes,
            status_code=r.status_code,
            media_type=r.headers.get("content-type"),
        )



from ..config import print_startup_info
from .base import LOG_CHAT, ForwardingBase


class AnyForwarding(ForwardingBase):
    def __init__(self, base_url: str, route_prefix: str, proxy=None):
        import httpx

        self.BASE_URL = base_url
        self.ROUTE_PREFIX = route_prefix
        self.client = httpx.AsyncClient(
            base_url=self.BASE_URL, proxies=proxy, http1=True, http2=False
        )
        print_startup_info(self.BASE_URL, self.ROUTE_PREFIX, [], "\\", LOG_CHAT)


def get_fwd_anything_objs():
    """获取extra路由转发对象"""
    from .settings import EXTRA_BASE_URL, EXTRA_ROUTE_PREFIX, PROXY

    extra_fwd_objs = []
    for base_url, route_prefix in zip(EXTRA_BASE_URL, EXTRA_ROUTE_PREFIX):
        extra_fwd_objs.append(AnyForwarding(base_url, route_prefix, PROXY))
    return extra_fwd_objs



from ..config import print_startup_info
from .base import OpenaiBase
from .settings import LOG_CHAT, OPENAI_API_KEY


class OpenaiForwarding(OpenaiBase):
    def __init__(self, base_url: str, route_prefix: str, proxy=None):
        import httpx

        self.BASE_URL = base_url
        self.ROUTE_PREFIX = route_prefix
        self.client = httpx.AsyncClient(
            base_url=self.BASE_URL, proxies=proxy, http1=True, http2=False
        )
        print_startup_info(
            self.BASE_URL,
            self.ROUTE_PREFIX,
            OPENAI_API_KEY,
            self._no_auth_mode,
            LOG_CHAT,
        )


def get_fwd_openai_style_objs():
    """获取openai风格路由转发对象"""
    from .settings import OPENAI_BASE_URL, OPENAI_ROUTE_PREFIX, PROXY

    fwd_objs = []
    for base_url, route_prefix in zip(OPENAI_BASE_URL, OPENAI_ROUTE_PREFIX):
        fwd_objs.append(OpenaiForwarding(base_url, route_prefix, PROXY))
    return fwd_objs
